/*
 * sudoku.c
 *
 * Copyright (c) 2005 Simon Budig  <simon@budig.de>
 */

#include <string.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk-pixbuf/gdk-pixdata.h>

#ifdef HILDON
#include <hildon-widgets/hildon-app.h>
#endif

#include "sudoku-puzzles.h"
#include "sudoku-board.h"

typedef struct _SudokuWindow SudokuWindow;

#define BOARD_SIZE  9

#define NUMBER_WIDTH 42
#define NUMBER_HEIGHT 42

#define NUMBER_MASK (0x0f)
#define PRESET_FLAG (1 << 4)
#define ERROR_FLAG  (1 << 5)
#define HINT_FLAG   (1 << 6)

#define NUMBER_TO_BIT(n)           (1 << (n))
#define BITFIELD_SET_NUMBER(f,n)   ((f) |= NUMBER_TO_BIT (n))
#define BITFIELD_CLEAR_NUMBER(f,n) ((f) &= ~NUMBER_TO_BIT (n))
#define BITFIELD_HAS_NUMBER(f,n)   ((f) & NUMBER_TO_BIT (n))

#define ROW_FROM_INDEX(i)   ((i) / BOARD_SIZE)
#define COL_FROM_INDEX(i)   ((i) % BOARD_SIZE)
#define BLOCK_FROM_INDEX(i) ((ROW_FROM_INDEX (i) / 3) * 3 +\
                             (COL_FROM_INDEX (i) / 3))

#define ROW_INDEX(r,i)   ((r) * BOARD_SIZE + i)
#define COL_INDEX(c,i)   ((c) + i * BOARD_SIZE)
#define BLOCK_INDEX(b,i) ((((b) / 3) * BOARD_SIZE + ((b) % 3)) * 3 +\
                           ((i) / 3) * BOARD_SIZE + ((i) % 3))

struct _SudokuWindow
{
  GtkWidget *appview;
  GtkWidget *drawing_area;
  GtkWidget *number_area;
  GtkWidget *menu;

  GdkPixmap *board_backing_pixmap;
  GdkPixmap *number_backing_pixmap;

  GdkPixbuf *image_data;
  GdkPixbuf *board_pixbuf;
  GdkPixbuf *numbers_pixbuf;

  gboolean   cursor_visible;
  gint       cursor_number;
  gint       cursor_x;
  gint       cursor_y;

  guint16   *board_state;
  guint16   *board_can_hold;

  guint16   *row_holds;
  guint16   *col_holds;
  guint16   *block_holds;

  guint16    row_errors;
  guint16    col_errors;
  guint16    block_errors;

  gboolean   error;
};


static void
sudoku_paint_number (SudokuWindow *swin, gint index)
{
  gint x = COL_FROM_INDEX (index);
  gint y = ROW_FROM_INDEX (index);
  gint number = swin->board_state[index];
  gboolean cursor = (x == swin->cursor_x && y == swin->cursor_y);

  x = x * NUMBER_WIDTH + 2 * (x / 3) + 3;
  y = y * NUMBER_HEIGHT + 2 * (y / 3) + 3;

  if (!swin->board_backing_pixmap)
    return;

  if (swin->cursor_visible && cursor)
    {
      gdk_draw_pixbuf (swin->board_backing_pixmap,
                       swin->drawing_area->style->black_gc,
                       swin->numbers_pixbuf,
                       0, NUMBER_HEIGHT,
                       x, y,
                       NUMBER_WIDTH - 1, NUMBER_HEIGHT - 1,
                       GDK_RGB_DITHER_MAX,
                       x, y);
    }
  else if (number & ERROR_FLAG)
    {
      gdk_draw_pixbuf (swin->board_backing_pixmap,
                       swin->drawing_area->style->black_gc,
                       swin->numbers_pixbuf,
                       0, 0,
                       x, y,
                       NUMBER_WIDTH - 1, NUMBER_HEIGHT - 1,
                       GDK_RGB_DITHER_MAX,
                       x, y);
    }
  else
    {
      gdk_draw_pixbuf (swin->board_backing_pixmap,
                       swin->drawing_area->style->black_gc,
                       swin->board_pixbuf,
                       x, y,
                       x, y,
                       NUMBER_WIDTH, NUMBER_HEIGHT,
                       GDK_RGB_DITHER_MAX,
                       x, y);
    }

  if ((number & NUMBER_MASK) > 0)
    {
      gdk_draw_pixbuf (swin->board_backing_pixmap,
                       swin->drawing_area->style->black_gc,
                       swin->numbers_pixbuf,
                       NUMBER_WIDTH * (number & NUMBER_MASK),
                       NUMBER_HEIGHT * (number & PRESET_FLAG ? 0 : 1),
                       x, y,
                       NUMBER_WIDTH - 1, NUMBER_HEIGHT - 1,
                       GDK_RGB_DITHER_MAX,
                       x, y);
    }

  gtk_widget_queue_draw_area (swin->drawing_area, x, y,
                              NUMBER_WIDTH - 1, NUMBER_WIDTH - 1);
}


static void
sudoku_paint_board (SudokuWindow *swin)
{
  gint i;

  for (i = 0; i < BOARD_SIZE * BOARD_SIZE; i++)
    sudoku_paint_number (swin, i);
}


static gboolean
sudoku_flag_errors (SudokuWindow *swin)
{
  gint i, row, col, block;

  for (i = 0; i < BOARD_SIZE * BOARD_SIZE; i++)
    {
      row = ROW_FROM_INDEX (i);
      col = COL_FROM_INDEX (i);
      block = BLOCK_FROM_INDEX (i);

      if (BITFIELD_HAS_NUMBER (swin->row_errors, row) ||
          BITFIELD_HAS_NUMBER (swin->col_errors, col) ||
          BITFIELD_HAS_NUMBER (swin->block_errors, block))
        {
          swin->board_state[i] |= ERROR_FLAG;
        }
      else
        {
          swin->board_state[i] &= ~ERROR_FLAG;
        }
    }

  sudoku_paint_board (swin);
  return swin->error;
}


static void
sudoku_update_state (SudokuWindow *swin,
                     gint          index)
{
  gint row, col, block, i, val;

  row = ROW_FROM_INDEX (index);
  col = COL_FROM_INDEX (index);
  block = BLOCK_FROM_INDEX (index);

  /* clear error bits for affected items */
  BITFIELD_CLEAR_NUMBER (swin->row_errors, row);
  BITFIELD_CLEAR_NUMBER (swin->col_errors, col);
  BITFIELD_CLEAR_NUMBER (swin->block_errors, block);

  /* recalculate the state */
  swin->row_holds[row] = 0;
  swin->col_holds[col] = 0;
  swin->block_holds[block] = 0;

  for (i = 0; i < BOARD_SIZE; i++)
    {
      val = swin->board_state [ROW_INDEX (row, i)] & NUMBER_MASK;
      if (val)
        {
          if (BITFIELD_HAS_NUMBER (swin->row_holds[row], val))
            BITFIELD_SET_NUMBER (swin->row_errors, row);
          BITFIELD_SET_NUMBER (swin->row_holds[row], val);
        }

      val = swin->board_state [COL_INDEX (col, i)] & NUMBER_MASK;
      if (val)
        {
          if (BITFIELD_HAS_NUMBER (swin->col_holds[col], val))
            BITFIELD_SET_NUMBER (swin->col_errors, col);
          BITFIELD_SET_NUMBER (swin->col_holds[col], val);
        }

      val = swin->board_state [BLOCK_INDEX (block, i)] & NUMBER_MASK;
      if (val)
        {
          if (BITFIELD_HAS_NUMBER (swin->block_holds[block], val))
            BITFIELD_SET_NUMBER (swin->block_errors, block);
          BITFIELD_SET_NUMBER (swin->block_holds[block], val);
        }
    }

  swin->error = swin->row_errors | swin->col_errors | swin->block_errors;
}


static gboolean
sudoku_place_number (SudokuWindow *swin,
                     gint          number,
                     gint          x,
                     gint          y)
{
  gint i, current;

  g_return_val_if_fail (0 <= x && x < BOARD_SIZE, FALSE);
  g_return_val_if_fail (0 <= y && y < BOARD_SIZE, FALSE);

  i = y * BOARD_SIZE + x;

  current = swin->board_state[i];

  if ((current & NUMBER_MASK) > 0 && (current & PRESET_FLAG))
    return FALSE;  /* preset value! */

  /* preserve error flag */
  number |= (current & ERROR_FLAG);

  swin->board_state[i] = number;

  sudoku_update_state (swin, i);
  sudoku_paint_number (swin, i);

  // sudoku_flag_errors (swin);

  return TRUE;
}


static void
sudoku_move_cursor (SudokuWindow *swin, gint x, gint y, gboolean visible)
{
  gint index;
  gboolean old_visible = swin->cursor_visible;
  gint old_cursor_x = swin->cursor_x;
  gint old_cursor_y = swin->cursor_y;

  swin->cursor_x = CLAMP (x, 0, 8);
  swin->cursor_y = CLAMP (y, 0, 8);
  swin->cursor_visible = visible;

  if (swin->cursor_visible)
    {
      index = swin->cursor_x + swin->cursor_y * BOARD_SIZE;

      sudoku_paint_number (swin, index);
    }

  if (old_visible)
    {
      index = old_cursor_x + old_cursor_y * BOARD_SIZE;

      sudoku_paint_number (swin, index);
    }
}


static void
sudoku_paint_number_area (SudokuWindow *swin)
{
  gint i, x, y;

  gdk_draw_pixbuf (swin->number_backing_pixmap,
                   swin->number_area->style->black_gc,
                   swin->board_pixbuf,
                   2, 2,
                   0, 0,
                   NUMBER_WIDTH * 2 + 1,
                   NUMBER_HEIGHT * 3 + 1,
                   GDK_RGB_DITHER_MAX,
                   0, 0);

  gdk_draw_pixbuf (swin->number_backing_pixmap,
                   swin->number_area->style->black_gc,
                   swin->board_pixbuf,
                   2, NUMBER_HEIGHT * 3 + 5,
                   0, NUMBER_WIDTH * 3 + 1,
                   NUMBER_WIDTH * 2 + 1,
                   NUMBER_HEIGHT * 2 + 1,
                   GDK_RGB_DITHER_MAX,
                   NUMBER_WIDTH * 3 + 1, 0);

  for (i = 0; i < 10; i++)
    {
      x = (i / 5) * NUMBER_WIDTH + 1;
      y = (i % 5) * NUMBER_HEIGHT + 1;

      if ((i + 1) % 10 == swin->cursor_number)
        {
          gdk_draw_pixbuf (swin->number_backing_pixmap,
                           swin->number_area->style->black_gc,
                           swin->numbers_pixbuf,
                           0, NUMBER_HEIGHT,
                           x, y,
                           NUMBER_WIDTH - 1, NUMBER_HEIGHT - 1,
                           GDK_RGB_DITHER_MAX,
                           x, y);
        }

      if (i < 9)
        {
          gdk_draw_pixbuf (swin->number_backing_pixmap,
                           swin->number_area->style->black_gc,
                           swin->numbers_pixbuf,
                           NUMBER_WIDTH * (i + 1),
                           NUMBER_HEIGHT,
                           x, y,
                           NUMBER_WIDTH - 1, NUMBER_HEIGHT - 1,
                           GDK_RGB_DITHER_MAX,
                           x, y);
        }
    }
}


static void
sudoku_clear_board (SudokuWindow *swin)
{
  gint i, current;

  for (i = 0; i < BOARD_SIZE * BOARD_SIZE; i++)
    {
      current = swin->board_state[i];
      if (current & PRESET_FLAG)
        current = PRESET_FLAG | (current & NUMBER_MASK);
      else
        current = 0;
      sudoku_place_number (swin, current,
                           COL_FROM_INDEX (i), ROW_FROM_INDEX (i));
    }

  sudoku_paint_board (swin);
}


static void
sudoku_setup_initial_board (SudokuWindow *swin)
{
  gint i, val, level;
  const gchar *pz;

  if (swin->board_state)
    g_free (swin->board_state);
  swin->board_state    = g_new0 (guint16, BOARD_SIZE * BOARD_SIZE);
  if (swin->board_can_hold)
    g_free (swin->board_can_hold);
  swin->board_can_hold = g_new0 (guint16, BOARD_SIZE * BOARD_SIZE);

  swin->row_holds   = g_new0 (guint16, BOARD_SIZE);
  swin->col_holds   = g_new0 (guint16, BOARD_SIZE);
  swin->block_holds = g_new0 (guint16, BOARD_SIZE);

  swin->error = 0;
  swin->row_errors = 0;
  swin->col_errors = 0;
  swin->block_errors = 0;

  pz = sudoku_puzzles [g_random_int_range (0, G_N_ELEMENTS (sudoku_puzzles))];

  level = pz[0] - '0';

  for (i=0; i < strlen (pz) - 2; i++)
    {
      val = pz[i+2];

      sudoku_place_number (swin,
                           (val >= 'a' ? 0 : ((val - '@') | PRESET_FLAG)),
                           COL_FROM_INDEX (i),
                           ROW_FROM_INDEX (i));
    }

}


static gboolean
sudoku_configure_event (GtkWidget         *widget,
                        GdkEventConfigure *event,
                        gpointer          *data)
{
  SudokuWindow *swin = (SudokuWindow *) data;
  GdkPixmap *new_pixmap;

  if (widget == swin->drawing_area)
    {
      if (swin->board_backing_pixmap)
        g_object_unref (G_OBJECT (swin->board_backing_pixmap));

      new_pixmap = gdk_pixmap_new (swin->drawing_area->window,
                                   swin->drawing_area->allocation.width,
                                   swin->drawing_area->allocation.height,
                                   -1);
      swin->board_backing_pixmap = new_pixmap;
      gdk_draw_pixbuf (swin->board_backing_pixmap,
                       swin->drawing_area->style->black_gc,
                       swin->board_pixbuf,
                       0, 0,
                       0, 0,
                       gdk_pixbuf_get_width (swin->board_pixbuf),
                       gdk_pixbuf_get_height (swin->board_pixbuf),
                       GDK_RGB_DITHER_MAX,
                       0, 0);
      sudoku_paint_board (swin);
    }

  if (widget == swin->number_area)
    {
      if (swin->number_backing_pixmap)
        g_object_unref (G_OBJECT (swin->number_backing_pixmap));

      new_pixmap = gdk_pixmap_new (swin->number_area->window,
                                   swin->number_area->allocation.width,
                                   swin->number_area->allocation.height,
                                   -1);

      swin->number_backing_pixmap = new_pixmap;
      sudoku_paint_number_area (swin);
    }

  return TRUE;
}


static gboolean
sudoku_expose_event (GtkWidget      *widget,
                     GdkEventExpose *event,
                     gpointer       *data)
{
  SudokuWindow *swin = (SudokuWindow *) data;

  if (widget == swin->drawing_area)
    {
      gdk_draw_pixmap (swin->drawing_area->window,
                       swin->drawing_area->style->black_gc,
                       swin->board_backing_pixmap,
                       event->area.x, event->area.y,
                       event->area.x, event->area.y,
                       event->area.width, event->area.height);
    }

  if (widget == swin->number_area)
    {
      gdk_draw_pixmap (swin->number_area->window,
                       swin->number_area->style->black_gc,
                       swin->number_backing_pixmap,
                       event->area.x, event->area.y,
                       event->area.x, event->area.y,
                       event->area.width, event->area.height);
    }

  return TRUE;
}


static gboolean
sudoku_key_press (GtkWidget   *widget,
                  GdkEventKey *event,
                  gpointer    *data)
{
  SudokuWindow *swin = (SudokuWindow *) data;
  gboolean      handled = TRUE;
  gint          number;

  switch (event->keyval)
    {
      case GDK_Up:
        sudoku_move_cursor (swin,
                            swin->cursor_x,
                            (swin->cursor_y + BOARD_SIZE - 1) % BOARD_SIZE,
                            TRUE);
        break;

      case GDK_Down:
        sudoku_move_cursor (swin,
                            swin->cursor_x,
                            (swin->cursor_y + 1) % BOARD_SIZE,
                            TRUE);
        break;

      case GDK_Left:
        sudoku_move_cursor (swin,
                            (swin->cursor_x + BOARD_SIZE - 1) % BOARD_SIZE,
                            swin->cursor_y,
                            TRUE);
        break;

      case GDK_Right:
        sudoku_move_cursor (swin,
                            (swin->cursor_x + 1) % BOARD_SIZE,
                            swin->cursor_y,
                            TRUE);
        break;

      case GDK_BackSpace:
      case GDK_Escape:
        sudoku_place_number (swin, 0, swin->cursor_x, swin->cursor_y);
        break;

      case GDK_ISO_Enter:
      case GDK_Return:
      case GDK_space:
        sudoku_place_number (swin, swin->cursor_number,
                             swin->cursor_x, swin->cursor_y);
        break;

      case GDK_plus:
      case GDK_F7:
        swin->cursor_number = (swin->cursor_number + 1) % 10;
        sudoku_paint_number_area (swin);
        gtk_widget_queue_draw (swin->number_area);
        break;

      case GDK_minus:
      case GDK_F8:
        swin->cursor_number = (swin->cursor_number + 9) % 10;
        sudoku_paint_number_area (swin);
        gtk_widget_queue_draw (swin->number_area);
        break;

      case GDK_h:
        sudoku_flag_errors (swin);
        break;

      default:
        number = gdk_keyval_to_unicode (event->keyval);

        if (number >= '0' && number <= '9')
          {
            swin->cursor_number = number - '0';
            sudoku_paint_number_area (swin);
            gtk_widget_queue_draw (swin->number_area);
            if (swin->cursor_visible)
              sudoku_place_number (swin, swin->cursor_number,
                                   swin->cursor_x, swin->cursor_y);

          }
        else
          {
            handled = FALSE;
          }

        break;
    }

  return handled;
}


static gboolean
sudoku_button_press (GtkWidget      *widget,
                     GdkEventButton *event,
                     gpointer       *data)
{
  SudokuWindow *swin = (SudokuWindow *) data;
  gint x = -1, y = -1;

  if (widget == swin->number_area)
    {
      x = event->x / NUMBER_WIDTH;
      y = event->y / NUMBER_HEIGHT;

      x = (y + x * 5 + 1) % 10;
      swin->cursor_number = x;

      sudoku_paint_number_area (swin);
      gtk_widget_queue_draw (widget);

      return TRUE;
    }

  if (event->x >= 3 && event->x <= 6 + 9 * NUMBER_WIDTH)
    {
      x = event->x - 2;
      if (x >= 3 * NUMBER_HEIGHT + 2) x--;
      if (x >= 3 * NUMBER_HEIGHT + 1) x--;
      if (x >= 6 * NUMBER_HEIGHT + 2) x--;
      if (x >= 6 * NUMBER_HEIGHT + 1) x--;

      if (x % NUMBER_HEIGHT != 0)
        x = x / NUMBER_WIDTH;
      else
        x = -1;
    }
  if (event->y >= 3 && event->y <= 6 + 9 * NUMBER_HEIGHT)
    {
      y = event->y - 2;
      if (y >= 3 * NUMBER_HEIGHT + 2) y--;
      if (y >= 3 * NUMBER_HEIGHT + 1) y--;
      if (y >= 6 * NUMBER_HEIGHT + 2) y--;
      if (y >= 6 * NUMBER_HEIGHT + 1) y--;

      if (y % NUMBER_HEIGHT != 0)
        y = y / NUMBER_HEIGHT;
      else
        y = -1;
    }

  if (x >= 0 && x < BOARD_SIZE && y >= 0 && y < BOARD_SIZE)
    {
      sudoku_place_number (swin, swin->cursor_number, x, y);
      sudoku_move_cursor (swin, x, y, FALSE);
    }

  return TRUE;
}


static void
sudoku_new_puzzle_cb (GtkMenuItem *menuitem,
                      gpointer     data)
{
  SudokuWindow *swin = (SudokuWindow *) data;

  sudoku_setup_initial_board (swin);
  sudoku_paint_board (swin);
}


static void
sudoku_clear_puzzle_cb (GtkMenuItem *menuitem,
                        gpointer     data)
{
  SudokuWindow *swin = (SudokuWindow *) data;

  sudoku_clear_board (swin);
}


static void
sudoku_check_puzzle_cb (GtkMenuItem *menuitem,
                        gpointer     data)
{
  SudokuWindow *swin = (SudokuWindow *) data;

  sudoku_flag_errors (swin);
}


static void
sudoku_build_menu (SudokuWindow *swin, GtkMenu *menu)
{
  GtkWidget *menu_item;

  menu_item = gtk_menu_item_new_with_label ("New Puzzle");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
  g_signal_connect (G_OBJECT (menu_item), "activate",
                    G_CALLBACK (sudoku_new_puzzle_cb), swin);

  menu_item = gtk_menu_item_new_with_label ("Restart");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
  g_signal_connect (G_OBJECT (menu_item), "activate",
                    G_CALLBACK (sudoku_clear_puzzle_cb), swin);

  menu_item = gtk_menu_item_new_with_label ("Check Puzzle");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
  g_signal_connect (G_OBJECT (menu_item), "activate",
                    G_CALLBACK (sudoku_check_puzzle_cb), swin);

  menu_item = gtk_menu_item_new_with_label ("Quit");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
  g_signal_connect (G_OBJECT (menu_item), "activate",
                    G_CALLBACK (gtk_main_quit), NULL);

  gtk_widget_show_all (GTK_WIDGET (menu));
}


static SudokuWindow *
sudoku_setup_ui ()
{
  SudokuWindow *swin;
  GtkWidget *window, *table;
  GdkPixdata pixdata;
  GError *error = NULL;

  swin = g_new0 (SudokuWindow, 1);

  swin->cursor_visible = FALSE;
  swin->cursor_number = 1;
  swin->cursor_x = 4;
  swin->cursor_y = 4;

  sudoku_setup_initial_board (swin);

  gdk_pixdata_deserialize (&pixdata, 54915, sudoku_board_pixdata, &error);
  if (error) g_printerr ("%s\n", error->message);
  swin->image_data = gdk_pixbuf_from_pixdata (&pixdata, FALSE, &error);
  if (error) g_printerr ("%s\n", error->message);

  swin->board_pixbuf = gdk_pixbuf_new_subpixbuf (swin->image_data,
                                                 0, 0, 387, 387);
  swin->numbers_pixbuf = gdk_pixbuf_new_subpixbuf (swin->image_data,
                                                   0, 387,
                                                   419, 2 * NUMBER_WIDTH - 1);

  window = g_object_new (
#ifdef HILDON
                         HILDON_TYPE_APP,
                         // "app-title", "Sudoku",
#else
                         GTK_TYPE_WINDOW,
#endif
                         "type", GTK_WINDOW_TOPLEVEL,
                         "events", GDK_KEY_PRESS_MASK,
                         "title", "Sudoku",
                         NULL);

#ifdef HILDON
  hildon_app_set_title (HILDON_APP (window), "Sudoku");

  swin->appview = g_object_new (HILDON_TYPE_APPVIEW,
                                // "title", "Sudoku",
                                // "fullscreen-key-allowed", TRUE,
                                "visible", TRUE,
                                "parent", window,
                                NULL);
  hildon_appview_set_title (HILDON_APPVIEW (swin->appview), "Sudoku");
  hildon_appview_set_fullscreen_key_allowed (HILDON_APPVIEW (swin->appview),
                                             TRUE);
  swin->menu = GTK_WIDGET (hildon_appview_get_menu (HILDON_APPVIEW (swin->appview)));
#else
  swin->appview = g_object_new (GTK_TYPE_VBOX,
                                "visible", TRUE,
                                "parent", window,
                                NULL);

  {
    GtkWidget *item, *menu_bar;

    menu_bar = g_object_new (GTK_TYPE_MENU_BAR,
                             "visible", TRUE,
                             "parent", swin->appview,
                             NULL);
    gtk_widget_show (menu_bar);

    item = gtk_menu_item_new_with_label ("Sudoku");
    gtk_menu_bar_append (GTK_MENU_BAR (menu_bar), item);
    gtk_widget_show (item);

    swin->menu = g_object_new (GTK_TYPE_MENU,
                               "visible", TRUE,
                               NULL);
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), swin->menu);
  }
#endif

  sudoku_build_menu (swin, GTK_MENU (swin->menu));

  table = g_object_new (GTK_TYPE_TABLE,
                        "n-columns", 2,
                        "n-rows", 1,
                        "homogeneous", FALSE,
                        "row-spacing", 0,
                        "parent", swin->appview,
#ifdef HILDON
                        "column-spacing", 0,
#else
                        "column-spacing", 30,
#endif
                        "visible", TRUE,
                        NULL);

  swin->drawing_area = g_object_new (GTK_TYPE_DRAWING_AREA,
                                     "events", GDK_EXPOSURE_MASK
                                               | GDK_BUTTON_PRESS_MASK,
                                     "visible", TRUE,
                                     "width-request", gdk_pixbuf_get_width (swin->board_pixbuf),
                                     "height-request", gdk_pixbuf_get_height (swin->board_pixbuf),
                                     NULL);

  gtk_table_attach (GTK_TABLE (table), swin->drawing_area,
                    0, 1, 0, 1, GTK_EXPAND, GTK_EXPAND, 0, 0);

  swin->number_area = g_object_new (GTK_TYPE_DRAWING_AREA,
                                    "events", GDK_EXPOSURE_MASK
                                              | GDK_BUTTON_PRESS_MASK,
                                    "visible", TRUE,
                                    "height-request", 5 * NUMBER_WIDTH + 1,
                                    "width-request", 2 * NUMBER_WIDTH + 1,
                                    NULL);

  gtk_table_attach (GTK_TABLE (table), swin->number_area,
                    1, 2, 0, 1, GTK_EXPAND, GTK_EXPAND, 0, 0);

  g_object_connect (G_OBJECT (window),
                    "signal::delete_event",
                           G_CALLBACK (gtk_main_quit), NULL,
                    "signal::key_press_event",
                           G_CALLBACK (sudoku_key_press), swin,
                    NULL);

  g_object_connect (G_OBJECT (swin->drawing_area),
                    "signal::configure_event",
                           G_CALLBACK (sudoku_configure_event), swin,
                    "signal::expose_event",
                           G_CALLBACK (sudoku_expose_event), swin,
                    "signal::button_press_event",
                           G_CALLBACK (sudoku_button_press), swin,
                    NULL);

  g_object_connect (G_OBJECT (swin->number_area),
                    "signal::configure_event",
                           G_CALLBACK (sudoku_configure_event), swin,
                    "signal::expose_event",
                           G_CALLBACK (sudoku_expose_event), swin,
                    "signal::button_press_event",
                           G_CALLBACK (sudoku_button_press), swin,
                    NULL);


  gtk_widget_show (window);

  return swin;
}


static void
sudoku_shut_down (SudokuWindow *swin)
{
  if (swin->board_state)
    g_free (swin->board_state);

  g_object_unref (swin->board_backing_pixmap);
  g_object_unref (swin->number_backing_pixmap);

  g_object_unref (swin->image_data);
  g_object_unref (swin->board_pixbuf);
  g_object_unref (swin->numbers_pixbuf);

  g_free (swin);

}


gint
main (gint argc, gchar *argv[])
{
  SudokuWindow *swin;
  gint i, j;

  gtk_init (&argc, &argv);
  swin = sudoku_setup_ui ();
  gtk_main ();
  sudoku_shut_down (swin);

  return 0;
}


